
function [Pose] = gPnLup_solver(struct)
% published:  2017-12-05, using Matlab 9.2.0.556344 (R2017a)
% Authors: Nora Horanyi, Zoltan Kato  
% [1]: Nora Horanyi, Zoltan Kato, Generalized Pose Estimation from Line Correspondences with Known Vertical Direction, 
% In Proceedings of International Conference on 3D Vision, Qingdao, China, 2017, IEEE
% All rights reserved.

% updated:  2020-04-27, using Matlab (R2018b)
% the code has been updated by hichem Abdellali

% input: struct
% struct.camdata(i).lines_2D_start = normalized coordinates of the image line segment's startpoint on the unit sphere
% struct.camdata(i).lines_2D_end = normalized coordinates of the image line segment's endpoint on the unit sphere
% struct.camdata(i).lines_3D_start = start point of the 3D line segment in the world frame
% struct.camdata(i).lines_3D_end = end point of the 3D line segment in the world frame
% struct.camdata(i).relpose_R = 3*3 rotation matrix of the relative position
% struct.camdata(j).relpose_t = 3*1 translation vector of the relative position
% struct.cambasic.angles = known vertical direction angles

% output: pose = 4*4 matrix estimated absolute pose (world coordinate system to camera) 

%Calculate normal direction vectors
for i=1:size(struct.camdata  ,2)
    nc=[];
    xs=struct.camdata(i).lines_2D_start;
    xe=struct.camdata(i).lines_2D_end;
    for j=1:size(xs,1)
        nc(:,j) = cross(xs(j,:),xe(j,:));
        nc(:,j) = nc(:,j)/norm(nc(:,j));
    end
    n{i,:}=nc;
end

%Calculate direction vectors
for i=1:size(struct.camdata  ,2)
    
    Ai_temp=struct.camdata(i).lines_3D_start;
    Bi_temp=struct.camdata(i).lines_3D_end;
    for j=1:size(Ai_temp,1)
        dir = (Bi_temp(j,:) - Ai_temp(j,:))';
        V(j,:) = dir/norm(dir);
    end
    Vw{i,:}=V';
    P{i,:}=Ai_temp';
    Q{i,:}=Bi_temp';
end

%Calculate Rv matrix based on known vertical direction
beta=struct.cambasic.angles(1);
gamma=struct.cambasic.angles(2);

Ry=[cos(beta) 0 sin(beta); 0 1 0; -sin(beta) 0 cos(beta)];
Rz=[cos(gamma), -sin(gamma), 0; sin(gamma), cos(gamma), 0; 0, 0, 1];

for i=1:size(struct.camdata  ,2)    
    Rv{i,:}=struct.camdata(i).relpose_R*Rz*Ry;
end

%Solve equation in the single unknown q
for i=1:size(n,1)
    for j=1:size(n{i},2)
        [a_temp,b_temp,c_temp]=abc_solver(n{i,:}(:,j), Rv{i}, Vw{i,:}(:,j));
        a(i,j) = 4*a_temp^2;
        b(i,j) = 6*a_temp*b_temp;
        c(i,j) = 4*a_temp*c_temp + 2*b_temp^2;
        d(i,j) = 2*b_temp*c_temp;
    end    
end

a=sum(cat(2, a(:)));
b=sum(cat(2, b(:)));
c=sum(cat(2, c(:)));
d=sum(cat(2, d(:)));

q = roots([a b c d]);
q = q(imag(q)==0); 

optimumrot_cw = Inf(3,3);
optimumpos_cw = Inf(3,1);


idx=1;
% Calculate Rx with the real solutions, substitute Rx into (3), solve the linear equation in t(pos_wc)
for i=1:size(q,1)
    
    Rx = [q(i)^2+1, 0, 0; 0, -q(i)^2+1, -2*q(i); 0, 2*q(i), -q(i)^2+1]/(1 + q(i)^2);
    for j=1:size(struct.camdata  ,2)
        Rr=struct.camdata(j).relpose_R;
        tr=struct.camdata(j).relpose_t;
        
        R_temp=Rr * Rz * Ry * Rx;
        [Mat]=matcalculation(size(n{j},2), n{j}, R_temp, P{j}, Rr, tr);
        Mats{j,:}=Mat;
    end
    
    Mat=cat(1, Mats{:});
    [~, ~, VMat] = svd(Mat);
    vec = VMat(:,4);
    vec = vec/vec(4);
    
    %Calculate rotation matrix
    rot_wc = Rz*Ry*Rx;
    
    pos_wc = vec(1:3);
    pos_cw = rot_wc'*(-pos_wc);
    
%%%%%%%%%%%%%%%%%%%%%   
 
    pose1 = [rot_wc' pos_cw; 0 0 0 1];
    pose_estimated = inv(pose1);
    
    % input for the backprojection fucntion of the line on the unit sphere
    ArrayData (:,1) =  struct.camdata(1).lines_2D_start(:,1);
    ArrayData (:,2) =  struct.camdata(1).lines_2D_start(:,2);
    ArrayData (:,3) =  struct.camdata(1).lines_2D_start(:,3);
    
    ArrayData (:,4) =  struct.camdata(1).lines_2D_end(:,1);
    ArrayData (:,5) =  struct.camdata(1).lines_2D_end(:,2);
    ArrayData (:,6) =  struct.camdata(1).lines_2D_end(:,3);
    
    ArrayData (:,7) =  P{1}(1,:)';
    ArrayData (:,8) =  P{1}(2,:)';
    ArrayData (:,9) =  P{1}(3,:)';
    
    ArrayData (:,10) = Q{1}(1,:)';
    ArrayData (:,11) = Q{1}(2,:)';
    ArrayData (:,12) = Q{1}(3,:)';
    
    ArrayData (:,13) = Vw{1}(1,:)';
    ArrayData (:,14) = Vw{1}(2,:)';
    ArrayData (:,15) = Vw{1}(3,:)';
    
    ArrayData (:,16) = n{1}(1,:)';
    ArrayData (:,17) = n{1}(2,:)';
    ArrayData (:,18) = n{1}(3,:)';
    
    
    % backprojection error on the unit sphere
    [line_error] =  Back_projection_error_new_fast(pose_estimated,ArrayData(:,1),ArrayData(:,2),ArrayData(:,3),...
        ArrayData(:,4),ArrayData(:,5),ArrayData(:,6),...
        ArrayData(:,7),ArrayData(:,8),ArrayData(:,9),...
        ArrayData(:,10),ArrayData(:,11),ArrayData(:,12),...
        ArrayData(:,13),ArrayData(:,14),ArrayData(:,15),struct.camdata(1).cameraType);
    
    %
    line_error_all(idx) = sum(line_error);
    pose_estimated_all{idx} = pose_estimated;
    
    % Note: we assume that the omni camera is looking to z+ direction 
    %g(1) parameter according to scaramuza spherical omni model is positive
    
        Zchecking1 = pose_estimated * [P{1};ones(1,size(P{1},2))] ;
        Zchecking2 = pose_estimated * [Q{1};ones(1,size(P{1},2))];

        if(min(Zchecking1(3,:))<0 || min(Zchecking2(3,:))<0)
            Znegative(idx) = 1;
        else
            Znegative(idx) = 0;
            
        end       

    idx = idx + 1;
    
end

line_error_all = line_error_all + 100000 * Znegative;
   
[~,Idx] = sort(line_error_all);
Pose =pose_estimated_all(Idx);

return

function [Mat]=matcalculation(n, nc, R, Pw, Rr, tr)
Mat = zeros(n, 4);
% tr= Rv * tr;
for i = 1:n
    nxi = nc(1,i);  nyi = nc(2,i);  nzi = nc(3,i);
    Pxi = Pw(1,i);        Pyi = Pw(2,i);        Pzi = Pw(3,i);
    
    % apply the constraint scalarproduct(Pi^c, ni^c) = 0
    Mat(2*i-1, 1) = nxi * Rr(1,1) + nyi * Rr(2,1) + nzi * Rr(3,1);
    Mat(2*i-1, 2) = nxi * Rr(1,2) + nyi * Rr(2,2) + nzi * Rr(3,2);
    Mat(2*i-1, 3) = nxi * Rr(1,3) + nyi * Rr(2,3) + nzi * Rr(3,3);
    Mat(2*i-1, 4) = nxi * (Pxi * R(1,1) + Pyi * R(1,2) + Pzi * R(1,3) )+ nyi * (Pxi * R(2,1) + Pyi * R(2,2) + Pzi * R(2,3) )+ nzi * (Pxi * R(3,1) + Pyi * R(3,2) + Pzi * R(3,3) ) + nxi*tr(1) + nyi*tr(2) + nzi * tr(3) ; %=(nc(:,i)'*tr)
end
return

function [a,b,c]=abc_solver(nc, Rv, Vw)
nxi = nc(1);        nyi = nc(2);        nzi = nc(3);
Vxi = Vw(1);        Vyi = Vw(2);        Vzi = Vw(3);

a = nxi * (Vxi * Rv(1,1) - Vyi * Rv(1,2) - Vzi * Rv(1,3)) + nyi * (Vxi * Rv(2,1) - Vyi * Rv(2,2) - Vzi * Rv(2,3)) + nzi * (Vxi * Rv(3,1) - Vyi * Rv(3,2) - Vzi * Rv(3,3));
b = 2 * (nxi * (Vyi * Rv(1,3)- Vzi * Rv(1,2)) + nyi * (Vyi * Rv(2,3)- Vzi * Rv(2,2)) + nzi * (Vyi * Rv(3,3)- Vzi * Rv(3,2)));
c = nxi * (Vxi * Rv(1,1) + Vyi * Rv(1,2) + Vzi * Rv(1,3)) + nyi * (Vxi * Rv(2,1) + Vyi * Rv(2,2) + Vzi * Rv(2,3)) + nzi * (Vxi * Rv(3,1) + Vyi * Rv(3,2) + Vzi * Rv(3,3));

return
